"""
An implementation of the Result object similar to Rust Results.
Result type annotations should be defined as Result[Ok_Type, Error_Type].
The error should always be a subclass of `exhausterr.errors.Error`.

Typical use case is the following:

>>> from exhausterr.result import Result, Ok, Err
>>> from exhausterr.error import Error

>>> class ZeroDivideErr(Error):
>>>     description: "Cannot divide by zero"
>>>     exception_cls = ZeroDivisionErro

>>> def divide(a: float, b: float) -> Result[float, ZeroDivideErr]:
>>>     if b == 0.:
>>>         return Err(ZeroDivideErr())
>>>     return Ok(a / b)

If your function can return multiple error types you can pass
them as normal unions, e.g., Result[float, ZeroDivideErr | OverflowErr]

If the happy path of your function returns None,
you can shorten Ok(None) to Ok(), and you may also use the
return annotation `NoneOr[YourErrorType]`,
which is equivalent to `Result[None, YourErrorType]`.


@date: 16.08.2024
@author: Baptiste Pestourie
"""

from __future__ import annotations

from enum import Enum, auto
from typing import (
    ClassVar,
    Final,
    Generic,
    Literal,
    NoReturn,
    Optional,
    TypeVar,
    Union,
    cast,
    Self,
)

from ._errors import Error

V = TypeVar("V")


# -- Type Variables --- #
R = TypeVar("R", bound=Optional[object], covariant=True)
E = TypeVar("E", bound=Optional[object], covariant=True)


class ExhausterrSentinel(Enum):
    """
    Defining a Sentinel object to represent unset results
    None is not an option because it's a valid value, and
    should be differentiated from the result being unset.
    """

    NOTSET = auto()


NotsetT = Literal[ExhausterrSentinel.NOTSET]

_NOTSET: Final = ExhausterrSentinel.NOTSET


class AbstractResult(Generic[R, E]):
    """
    Base class for result objects Ok and Err.
    Abstract-ness is not enforced, you should however instanciate
    Results using Ok() and Err() unless you plan to extend functionalities.

    Part of the public interface for the following purposes:
        * Allow building custom classes
        * Use with isinstance() checks.

    For type hinting, use Result for Ok() and Err() objects
    rather than AbstractResult.
    """

    __slots__ = ("value", "error")
    value: R
    error: E

    def __bool__(self) -> bool:
        """
        Returns
        -------
        bool
            False for a Result that contains an error,
            True otherwise
        """
        return self.error is ExhausterrSentinel.NOTSET

    def unwrap(self) -> R:
        """
        Raises an exception ('panics') if this result
        is an error, otherwise, returns the result.
        Overriden by the subclasses, Ok() and Err().
        """
        raise RuntimeError(
            "This result object is empty and has never been set."
            "You should not call AbstractResult directly"
        )

    def unwrap_or_none(self) -> R | None:
        """
        Returns None if this Result is an error, otherwise, returns the result.
        Overriden by the subclasses, Ok() and Err().
        """
        raise RuntimeError(
            "This result object is empty and has never been set."
            "You should not call AbstractResult directly"
        )

    def unwrap_or_self(self) -> R | Self:
        """
        Returns the inner value if this Result is OK,
        otherwise returns itself.
        Intended to be use to forward ('reraise') errors easily, e.g,
        assuming func() returns a Result, and we just want to re-return
        the error if it errors out:
        >>> if not (val := func().unwrap_or_self()):
        >>>      return val
        Mind that the wrapped result value should always be truthy for this pattern to work !
        Overriden by the subclasses, Ok() and Err().
        """
        raise RuntimeError(
            "This result object is empty and has never been set."
            "You should not call AbstractResult directly"
        )


class Ok(AbstractResult[R, NotsetT], Generic[R]):
    """
    A successful result.

    """

    __match_args__: ClassVar[tuple[str, ...]] = ("value",)
    value: R

    def __init__(self, value: R = cast(R, None)) -> None:
        """
        Parameters
        ----------
        value: R
            The returned value
        """
        self.value = value
        self.error = _NOTSET

    def __repr__(self) -> str:
        """
        Shows only the internal value. Should rebuild the object when fed into eval().
        """
        return f"Ok({self.value!r})"

    def __bool__(self) -> Literal[True]:
        """
        Returns
        -------
        bool
            False for a Result that contains an error,
            True otherwise
        """
        return True

    def __eq__(self, other: object) -> bool:
        """
        Makes the comparison of two `Ok` objects equivalent to
        compare their wrapped values.
        """
        if not isinstance(other, AbstractResult):
            return NotImplemented
        return other.error is _NOTSET and self.value == other.value

    def unwrap(self) -> R:
        """
        Returns
        -------
        R
            Inner result value.
        """
        return self.value

    def unwrap_or_none(self) -> R:
        """
        Returns
        -------
        R
            Inner result value.
        """
        return self.value

    def unwrap_or_self(self) -> R:
        """
        Returns
        -------
        R
           Inner result value.
        """
        return self.value

    def __str__(self) -> str:
        """
        A human-friendly wrapper around the internal value __str__
        """
        return f"Ok({self.value})"

    @property
    def inner(self) -> R:
        """
        A method shared with `Err`. Returns the inner value, useful
        to iterate over multiple Results when one just wants the wrapped error or value
        regardless of whether it is an `Ok` or an `Err`.
        """
        return self.value


class Err(AbstractResult[NotsetT, E], Generic[E]):
    """
    An error result.
    """

    __slots__ = AbstractResult.__slots__ + ("_exception_cls",)

    __match_args__: ClassVar[tuple[str, ...]] = ("error",)
    error: E

    def __init__(
        self,
        error: E = cast(E, None),
        *,
        exception_cls: type[Exception] | None = None,
    ) -> None:
        """
        Parameters
        ----------
        value: R
            The returned error
        """
        self.error = error
        self.value = _NOTSET
        self._exception_cls = exception_cls

    def __repr__(self) -> str:
        """
        Shows only the internal error. Should rebuild the object when fed into eval().
        """
        return f"Err({self.error!r})"

    def __eq__(self, other: object) -> bool:
        """
        Makes the comparison of two `Err` objects equivalent to
        compare their wrapped values.
        """
        if not isinstance(other, AbstractResult):
            return NotImplemented
        return other.error is not _NOTSET and self.error == other.error

    def __bool__(self) -> Literal[False]:
        """
        Returns
        -------
        bool
            False for a Result that contains an error,
            True otherwise
        """
        return False

    def unwrap(self) -> NoReturn:
        """
        Returns
        -------
        NoReturns
           Always raises.
        """
        err = self.error
        if isinstance(err, Error):
            err.throw()
        raise Exception(err)

    def unwrap_or_none(self) -> None:
        """
        Returns
        -------
        None
            Always None
        """
        return None

    def unwrap_or_self(self) -> Self:
        """
        Returns
        -------
        Self
           Itself.
        """
        return self

    def __str__(self) -> str:
        """
        A human-friendly wrapper around the internal value __str__
        """
        return f"Err({self.error})"

    @property
    def inner(self) -> E:
        """
        A method shared with `Err`. Returns the inner value, useful
        to iterate over multiple Results when one just wants the wrapped error or value
        regardless of whether it is an `Ok` or an `Err`.
        """
        return self.error


# --- Result type hints --- #
Result = Union[Ok[R], Err[E]]
NoneOr = Union[Ok[None], Err[E]]



"""
Test suite for results objects.

@date: 16.08.2024
@author: Baptiste Pestourie
"""

from typing import assert_never

import pytest
from exhausterr import AbstractResult, Result, Ok, Err
from test_errors import BasicError


def fails_on_true(should_fail: bool) -> Result[bool, BasicError]:
    if should_fail:
        return Err(BasicError())
    return Ok(True)


def test_basic_result() -> None:
    """
    Test basic result.
    """
    result = fails_on_true(False)
    assert isinstance(result, AbstractResult)
    # checking __bool__
    assert result

    result = fails_on_true(True)
    assert isinstance(result, AbstractResult)
    assert isinstance(result.error, BasicError)
    assert not result, repr((result._ret, result.err))


@pytest.mark.parametrize("successful", [True, False])
def test_result_matching(successful: bool) -> None:
    """
    Test basic result.
    """
    result = fails_on_true(not successful)
    match result:
        case Ok(value):
            assert successful
            assert value is True
        case Err(error):
            assert not successful
            assert isinstance(error, BasicError)
        case _ as x:
            assert_never(x)
            assert False, "Unexpected result type"

    match result:
        case Ok(value):
            assert successful
            assert value is True
        case Err(error):
            assert not successful
            assert isinstance(error, BasicError)
        case _ as x:
            assert_never(x)
            assert False, "Unexpected result type"


def test_result_unwrap() -> None:
    """
    Checks that unwrap() raises an exception when the result is an error
    and returns the value when it is not.
    """
    fails = fails_on_true(True)
    with pytest.raises(Exception):
        fails.unwrap()

    assert fails_on_true(False).unwrap() is True
    fails = fails_on_true(True)
    with pytest.raises(Exception):
        fails.unwrap()

    assert fails_on_true(False).unwrap() is True


def test_result_repr_evals() -> None:
    an_ok_int = Ok(5)
    a_string_error = Err("some error")
    assert eval(repr(an_ok_int)) == an_ok_int
    assert eval(repr(a_string_error)) == a_string_error



"""
Test suite for error module.

@date: 16.08.2024
@author: Baptiste Pestourie
"""

from typing import TypedDict, NotRequired
from dataclasses import dataclass
import pytest
from exhausterr import Error
from exhausterr.type_utils import order_typed_dict_keys


def test_order_typed_dict_keys() -> None:
    """
    Checks that TypedDict key ordering returns the correct output when
    only required keys are given.
    """

    class _TypedDictTest(TypedDict):
        a: str
        b: str
        c: str

    # Note: looping multiple times as ordering from TypedDict are random
    for _ in range(10):
        assert list(order_typed_dict_keys(_TypedDictTest)) == ["a", "b", "c"]


def test_order_typed_dict_keys_mixed() -> None:
    """
    Checks that TypedDict key ordering returns the correct output when
    oa mix of required and optional keys is given
    """

    class _TypedDictTest(TypedDict):
        a: NotRequired[str]
        b: str
        c: str

    # Note: looping multiple times as ordering from TypedDict are random
    for _ in range(10):
        assert list(order_typed_dict_keys(_TypedDictTest)) == ["b", "c", "a"]


def test_order_typed_dict_keys_inherited() -> None:
    """
    Checks that TypedDict key ordering returns the correct output when
    the TypedDict inherits from another one.
    """

    class _BaseDict(TypedDict):
        a: NotRequired[str]
        b: str

    class _ChildDict(_BaseDict):
        c: NotRequired[str]
        d: str

    # Note: looping multiple times as ordering from TypedDict are random
    for _ in range(10):
        assert list(order_typed_dict_keys(_ChildDict)) == ["b", "d", "a", "c"]


class BasicError(Error):
    """
    A basic error that does not introduce any additional
    feature compared to a raw Error
    """

    description = "A basic error"
    exception_cls = ValueError


def test_error_parent_cls() -> None:
    """
    Checks that Error can be instanciated properly.
    Check that error without argument can be declared without @dataclass properly.
    """
    err = BasicError()
    assert BasicError.__match_args__ == ()
    assert str(err) == BasicError.description
    assert err.args == {}


def test_error_default_description() -> None:
    """
    Checks that description for errors that do not provide a custom
    one is properly formatted.
    """
    err = BasicError()
    assert err.description == "A basic error"


def test_error_throw() -> None:
    """
    Checks that Error.throw() raises the correct exception
    """
    err = BasicError()
    with pytest.raises(ValueError):
        err.throw()
    try:
        err.throw()
    except ValueError as e:
        assert str(e) == "A basic error"


@pytest.mark.parametrize("notes", (["note1", "note2"],))
def test_error_notes(notes: list[str]) -> None:
    """
    Checks that Error.notes() returns the correct notes
    """
    err = BasicError()
    err.add_notes(*notes)
    for note in notes:
        assert note in err._notes

    try:
        err.throw()
    except ValueError as e:
        for note in notes:
            assert note in e.__notes__


@dataclass
class ErrorWithArgs(Error):
    """
    A basic error that does not introduce any additional
    feature compared to a raw Error
    """

    description = "a: {a}, b: {b}"
    exception_cls = ValueError

    # Argument definition for this error

    a: int
    b: float


def test_error_meta_creates_match_args() -> None:
    """
    Checks that the matcha rguments are propely formatted
    by the metaclass when defining custom arguments
    """
    assert ErrorWithArgs.__match_args__ == ("a", "b")


def test_error_with_arguments_matching() -> None:
    """
    Checks that the match arguments are propely formatted
    by the metaclass when defining custom arguments
    """
    err = ErrorWithArgs(a=1, b=2.0)

    match err:
        case BasicError():
            assert False, "Matched wrongly as BasicError"

        case ErrorWithArgs(a, b):
            assert a == 1
            assert b == 2.0

        case _:
            assert False, "Error did not match"

    # now matching with keyword args
    match err:
        case BasicError():
            assert False, "Matched wrongly as BasicError"

        case ErrorWithArgs(a=a, b=b):
            assert a == 1
            assert b == 2.0

        case _:
            assert False, "Error did not match"


class ErrorWihFormatting(ErrorWithArgs):
    """
    A basic error that does not introduce any additional
    feature compared to a raw Error
    """

    default_description = "defaulted"
    description_header = "{a},{b}:"
    exception_cls = ValueError


def test_error_formatting() -> None:
    """
    Checks that errors using string expected format arguments
    are formatted properly
    """
    err = ErrorWihFormatting(a=1, b=2.0)
    assert str(err) == ("a: 1, b: 2.0")



# SPDX-FileCopyrightText: 2024-present Hedwyn <bpestourie@gmail.com>
#
# SPDX-License-Identifier: MIT
from ._results import Result, Ok, Err, NoneOr, AbstractResult
from ._errors import Error
from ._iterators import resultify, result_mapper, result_reducer, chain_results
from ._builtin import (
    safe_int_divide,
    safe_divide,
    safe_getattr,
    safe_getitem,
    ZeroDivision,
    BadKey,
    BadAttribute,
)

__all__ = [
    "AbstractResult",
    "Result",
    "Ok",
    "Err",
    "NoneOr",
    "Error",
    "resultify",
    "result_mapper",
    "result_reducer",
    "chain_results",
    "safe_int_divide",
    "safe_divide",
    "safe_getattr",
    "safe_getitem",
    "ZeroDivision",
    "BadKey",
    "BadAttribute",
]



"""
Some convenience iteration and composition tools around Results.

@date: 14.10.2024
@author: Baptiste Pestourie
"""

from __future__ import annotations
from typing import (
    Iterable,
    cast,
    assert_never,
    overload,
    Iterator,
    TypeVar,
    Callable,
    Any,
    reveal_type,
)

from ._results import Ok, Result, Err, NoneOr
from ._errors import Error

R = TypeVar("R")

# --- Bunch of type variables for composition ---
_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")
_D = TypeVar("_D")
_E = TypeVar("_E")
T = TypeVar("T")


def resultify(values: Iterable[R]) -> Iterator[Ok[R]]:
    """
    Maps bunch of values to Ok() results.
    """
    return map(Ok, values)


@overload
def result_mapper(
    collection: Iterable[_A], func1: Callable[[_A], Result[_B, Error]], /
) -> Iterable[Result[_B, Error]]: ...


@overload
def result_mapper(
    collection: Iterable[_A],
    func1: Callable[[_A], Result[_B, Error]],
    func2: Callable[[_B], Result[_C, Error]],
    /,
) -> Iterable[Result[_C, Error]]: ...


@overload
def result_mapper(
    collection: Iterable[_A],
    func1: Callable[[_A], Result[_B, Error]],
    func2: Callable[[_B], Result[_C, Error]],
    func3: Callable[[_C], Result[_D, Error]],
    /,
) -> Iterable[Result[_D, Error]]: ...


@overload
def result_mapper(
    collection: Iterable[_A],
    func1: Callable[[_A], Result[_B, Error]],
    func2: Callable[[_B], Result[_C, Error]],
    func3: Callable[[_C], Result[_D, Error]],
    func4: Callable[[_D], Result[_E, Error]],
    /,
) -> Iterable[Result[_E, Error]]: ...


def result_mapper(
    collection: Iterable[Any], *funcs: Callable[[Any], Result[Any, Error]]
) -> Iterable[Result[Any, Error]]:
    """
    Composes functions together and used the composed function to map values from a collection.

    Yields
    -------
    For every value `v` in `collection`
    (func_n => func_n-1 => ... => func_1)(v)
    (`=>` denoting composition), when all func return Ok.
    For any k in n for which func_n would return an Err(),
    skips all the next funcs and returns the Err for that k.

    Parameters
    ----------
    collection: Iterable[Any]
        The values to map.

    funcs: Iterable[Callable[[A], Result[B, Error]]]
        Functions to compose together
    """
    for item in collection:
        _result: Result[Any, Error] = Ok(item)
        _error: Error | None = None
        for func in funcs:
            match _result:
                case Ok(value):
                    _result = func(value)
                case Err(error):
                    _error = error
                    break

        if _error is not None:
            yield Err(_error)
        else:
            yield _result


@overload
def result_reducer(
    *results: Result[Any, Error],
    reducer: None,
    initializer: None,
) -> NoneOr[Error]: ...


@overload
def result_reducer(
    *results: Result[Any, Error],
    reducer: Callable[[T, Any], T],
    initializer: T,
) -> Result[T, Error]: ...


def result_reducer(
    *results: Result[Any, Error],
    reducer: Callable[[T, Any], T] | None = None,
    initializer: T | None = None,
) -> Result[Any, Error]:
    """
    Similar to functools.reduce, except it operates
    on Result objects and stops on the first error encountered.
    If getting any error while chaining, returns the error.
    Otherwise, reduces the value wrapped by `Ok`.
    Reduction is defined by:
    reduce(Seq[n + 1]) = reduce(reduce(Seq[0..n]), Seq[n + 1])
    with value for the first call being `initializer`.

    Examples
    --------
    >>> result_reducer(
        [Ok(2), Ok(3), Ok(5)],
        initializer=[],
        reducer=list.append
    )
    Ok([2,3,5])

    >>> result_reducer(
        [Ok(2), Ok(3), Ok(4)],
        initializer=0,
        reducer=sum
    )
    Ok(10)

    >>> result_reducer(
        [Ok(2), Err(3), Ok(4)],
        initializer=0,
        reducer=sum
    )
    Err(3)
    """
    reduced_value = cast(T, initializer)
    for r in results:
        match r:
            case Ok(val):
                if reducer is not None:
                    reduced_value = reducer(reduced_value, val)
            case Err(error):
                return Err(error)

            case _ as unreachable:
                assert_never(unreachable)

    return Ok(reduced_value)


@overload
def chain_results(
    *funcs: Callable[[], Result[Any, Error]],
    initializer: None,
    reducer: None,
) -> NoneOr[Error]: ...


@overload
def chain_results(
    *funcs: Callable[[], Result[T, Error]],
    initializer: T,
    reducer: Callable[[T, Any], T],
) -> Result[T, Error]: ...


def chain_results(
    *funcs: Callable[[], Result[Any, Error]],
    initializer: T | None = None,
    reducer: Callable[[T, Any], T] | None = None,
) -> Result[Any, Error]:
    """
    Similar to `result_reducer`, but takes a functions that returns
    results instead of results.
    Calls the function in `funcs` until an error is obtained
    or all or them are consumed.
    Output is the same as what `result_reducer` would
    give on the functions return values.

    See also
    --------
    result_reducer
    """
    reduced_value = cast(T, initializer)
    for f in funcs:
        res = f()
        match res:
            case Ok(val):
                if reducer:
                    reduced_value = reducer(reduced_value, val)
            case Err(error):
                return Err(error)

            case _ as unreachable:
                assert_never(unreachable)

    return Ok(reduced_value)



"""
Base layer for error types.

@author: Baptiste Pestourie
@date: 16.08.2024
"""

from __future__ import annotations
from typing import (
    NoReturn,
    Any,
    ClassVar,
)
from functools import cached_property
from dataclasses import is_dataclass, dataclass, field, asdict


@dataclass
class Error:
    """
    Base representation for an error.

    If you want your error to support arguments, decorate it
    with @dataclass and declare your arguments as fields.
    Private variables (starting with '_') are excluded from
    args property.


    Class parameters
    ----------------
    description: str
        The description for the error, with optional format arguments.
        F-string arguments should match argument names.

    exception_cls: type[Exception]
        When raising that error as an execption,
        the exception class that will be used.
    """

    description: ClassVar[str | None] = None
    exception_cls: ClassVar[type[Exception]] = Exception
    _notes: list[str] = field(
        init=False, compare=False, hash=False, default_factory=list
    )

    @cached_property
    def _is_dataclass(self) -> bool:
        """
        Returns
        -------
        bool
            Whether this object is a dataclass
        """
        return is_dataclass(self)

    @property
    def args(self) -> dict[str, Any]:
        """
        Returns
        -------
        dict[str, Any]
            Dict containing the arguments and their current values
        """
        return {name: value for name, value in asdict(self).items() if name[0] != "_"}

    def add_notes(self, *notes: str) -> None:
        """
        Adds a note to give more details on this error object.
        """
        self._notes.extend(notes)

    def __str__(self) -> str:
        """
        Returns
        -------
        str
            The complete error description
        """
        if self.description is None:
            return ""
        return self.description.format(**self.args)

    def throw(self) -> NoReturn:
        """
        Raises an exception from this error.
        Uses the exception class `exception_cls` defined at the class-level.
        """
        exc = self.exception_cls(self.description)
        for note in self._notes:
            exc.add_note(note)
        raise exc



"""
Test suite for the error wrappers on as alternative to builtin primitives.

@date: 15.10.2024
@author: Baptiste Pestourie
"""

from __future__ import annotations
from dataclasses import dataclass
import pytest
from exhausterr import (
    Ok,
    Err,
    safe_int_divide,
    safe_divide,
    safe_getattr,
    safe_getitem,
    ZeroDivision,
    BadKey,
    BadAttribute,
)


@pytest.mark.parametrize("a,b", [(5, 0), (5, 2)])
def test_safe_divide(a: float, b: float) -> None:
    match safe_divide(a, b):
        case Ok(value):
            assert b != 0
            assert value == a / b

        case Err(err):
            assert isinstance(err, ZeroDivision)


@pytest.mark.parametrize("a,b", [(5, 0), (5, 2)])
def test_safe_int_divide(a: int, b: int) -> None:
    match safe_int_divide(a, b):
        case Ok(value):
            assert b != 0
            assert value == a // b

        case Err(err):
            assert isinstance(err, ZeroDivision)


def test_safe_getattr() -> None:
    @dataclass
    class Dummy:
        a: int = 1
        b: int = 2

    dummy = Dummy()
    for attr in ["a", "b", "c"]:
        match safe_getattr(dummy, attr):
            case Ok(value):
                assert attr in ("a", "b")
                assert value == getattr(dummy, attr)

            case Err(err):
                assert attr == "c"
                assert isinstance(err, BadAttribute)


def test_safe_getitem() -> None:
    dummy = {"a": 1, "b": 2}
    for attr in ["a", "b", "c"]:
        match safe_getitem(dummy, attr):
            case Ok(value):
                assert attr in ("a", "b")
                assert value == dummy[attr]

            case Err(err):
                assert attr == "c"
                assert isinstance(err, BadKey)



"""
Type definitions for better type annotations.

@date: 16.08.2024
@author: Baptiste Pestourie
"""

from typing import Protocol, Iterator
import itertools


class TypedDictDefinition(Protocol):
    """
    Patches the absence of public interface metaclass definition
    for TypedDict (_TypedDictMet being private).
    Functions taking a TypedDict definition as argument can simply not
    type properly their arguments.
    """

    __annotations__: dict[str, type]
    __optional_keys__: frozenset[str]
    __required_keys__: frozenset[str]


def order_typed_dict_keys(typed_dict: TypedDictDefinition) -> Iterator[str]:
    """
    Allows ordering the keys of a typed dict in declaration order,
    as __required_keys__ and __optional_keys__ are not ordered.

    Yields
    ------
    str
        Required keys in declaration order, followed by optional keys in declaration order.
        Note that this means that optional keys declared before required one will still
        be put after in this ordering.
    """
    required_keys = typed_dict.__required_keys__

    required_keys_ordered: list[str] = []
    optional_keys_ordered: list[str] = []

    for annotation in typed_dict.__annotations__.keys():
        container = (
            required_keys_ordered
            if annotation in required_keys
            else optional_keys_ordered
        )
        container.append(annotation)

    return itertools.chain(required_keys_ordered, optional_keys_ordered)


"""
Some wrapper on builtin Python exceptions and primitives to allow using
them in a railway style.
Creates `Error` alter egos for some of the builtin Pythjn exceptions.
Provides equivalent primitives to some dunder-calling functions that return
`Result` instead of raising exceptions.
Note: due to Python using `Error` as suffix on builtin exceptions, we refrain from using
that keyword in the Error class names.


@date: 20.09.2024
@author: Baptiste Pestourie
"""

from __future__ import annotations
from typing import ClassVar, Final, Mapping
from dataclasses import dataclass
from ._results import Result, Ok, Err
from ._errors import Error

_NO_DEFAULT: Final[object] = object()


@dataclass
class BadAttribute(Error):
    """
    The error equivalent of built-in exception AttributeError.
    Takes as parameter the object on which the attribute access failed
    as well as the attribute name.
    """

    exception_cls: ClassVar[type[Exception]] = AttributeError
    description: ClassVar[str | None] = "Object {obj} has no attribute {attr}"
    obj: object
    attr: str


@dataclass
class BadKey(Error):
    """
    The error equivalent of built-in exception KeyError.
    Takes as paramter the mapping on which the key access failed
    as well as the attribute name.
    """

    exception_cls: ClassVar[type[Exception]] = AttributeError
    description: ClassVar[str | None] = "Mapping {mapping} has no attribute {key}"
    mapping: Mapping[object, object]
    key: object


class ZeroDivision(Error):
    """
    The error equivalent of built-in exception KeyError.
    Takes as paramter the mapping on which the key access failed
    as well as the attribute name.
    """

    exception_cls: ClassVar[type[Exception]] = ZeroDivisionError
    description: ClassVar[str | None] = "Cannot divide by zero"


def safe_int_divide(a: int, b: int) -> Result[int, ZeroDivision]:
    """
    Railway variant for division for integers,
    returns  Ok(`a // b`) if `b != 0` else `Err(ZeroDivisionErr())`
    """
    if b == 0:
        return Err(ZeroDivision())
    return Ok(a // b)


def safe_divide(a: float, b: float) -> Result[float, ZeroDivision]:
    """
    Railway variant for `__truediv__` for integers,
    returns  Ok(`a // b`) if `b != 0` else `Err(ZeroDivisionErr())`
    """
    if b == 0:
        return Err(ZeroDivision())
    return Ok(a / b)


def safe_getattr(
    obj: object, attr: str, default: object = _NO_DEFAULT
) -> Result[object, BadAttribute]:
    """
    Railway variant for `getattr`.
    Unlike `getattr`, `default` can also be passed as keyword argument
    (instead of positional only).

    Returns
    -------
    Result[Any, AttributeErr]
        The attribute value if it exists.
        AttributeErr Otherwise.
    """

    found = getattr(obj, attr, default)
    if found != _NO_DEFAULT:
        return Ok(found)
    return Err(BadAttribute(obj, attr))


def safe_getitem(
    obj: Mapping[object, object], key: str, default: object = _NO_DEFAULT
) -> Result[object, BadKey]:
    """
    Railway variant for `__getitem__`, triggered by the subscript operator ([]).
    Supports default arguments as well.

    Returns
    -------
    Result[Any, KeyErr]
        The attribute value if it exists.
        AttributeErr Otherwise.
    """
    # Note: using contextlib.suppress is more elegant but slower
    try:
        return Ok(obj[key])
    except KeyError:
        if default != _NO_DEFAULT:
            return Ok(default)
        return Err(BadKey(obj, key))



from exhausterr import Result, Error
from typing import reveal_type

def check_result(result: Result[int, Error]) -> None:
    """
    Demonstrates that type narrowing is properly performed
    when using `if` statements instead of `match`.
    Running 'if result: ...' will narrow the result to Ok(...)
    wihin `if` scope and to Err(...) in the `else` scope.
    """
    if result:
        # revealed type is Ok[int]
        reveal_type(result)
    else:
        # revealed type is Err[Error]
        reveal_type(result)


print("Note: this example is meant to be run by mypy, not to be executed")



"""
A naive example on how to use Result and Error
"""

import random
from enum import Enum, auto
from typing import assert_never, reveal_type

from exhausterr import Err, Error, Ok, Result


class CoinTossResult(Enum):
    HEADS = auto()
    TAILS = auto()


class LandedOnEdge(Error):
    pass


class DownTheGutter(Error):
    pass


def toss_a_coin() -> Result[CoinTossResult, LandedOnEdge | DownTheGutter]:
    """
    Plays heads or tails... with a few twists !
    """
    rng = random.random()
    if rng < 0.1:
        return Err(DownTheGutter())

    if rng < 0.2:
        # jeez, we landed on an edge
        return Err(LandedOnEdge())

    result = CoinTossResult.HEADS if rng < 0.6 else CoinTossResult.TAILS
    return Ok(result)


def play() -> None:
    """
    Tosses a coin and informs the player about the result
    """
    res = toss_a_coin()
    match res:
        case Ok(value=coin):
            reveal_type(coin)
            print(
                f"Got {coin.name.lower()}, you {'won' if coin == CoinTossResult.HEADS else 'lost'} !"
            )

        case Err(error=err):
            reveal_type(err)
            # Something went wrong.. but what
            match err:
                case LandedOnEdge():
                    print("You landed on an edge ! Let's flip the coin again.")

                case DownTheGutter():
                    print("Ops, you lost a coin, let's get another one !")

                case _ as unreachable:
                    assert_never(unreachable)

        case _ as unreachable:
            assert_never(unreachable)


for _ in range(100):
    play()



"""
Test suite for Result mapping primitives

@date: 14.10.2024
@author: Baptiste Pestourie
"""

from __future__ import annotations
from functools import partial
from exhausterr import resultify, result_mapper, Ok, Error, Err, Result, result_reducer, chain_results


class ZeroDivision(Error):
    description = "Cannot divide by zero"


def test_resultify() -> None:
    """
    Sanity checks for resultify().
    """
    values = range(100)
    for result in resultify(values):
        assert isinstance(result, Ok)


# test primitives
def invert(value: float) -> Result[float, ZeroDivision]:
    if value == 0:
        return Err(ZeroDivision())
    return Ok(1 / value)


def subtract_to_one_and_invert(value: float) -> Result[float, ZeroDivision]:
    if value == 1.0:
        return Err(ZeroDivision())
    return Ok(1 / (1 - value))


def test_result_mapper_single_function() -> None:
    """
    Tests the result mapper with a single map function
    """
    values = range(5)
    mapped = list(result_mapper(values, invert))
    assert mapped[1:5] == [Ok(1.0), Ok(1 / 2), Ok(1 / 3), Ok(1 / 4)]
    assert not mapped[0]
    assert type(mapped[0].error) is ZeroDivision


def test_result_mapper_multiple_function() -> None:
    """
    Tests the result mapper with multiple functions composed
    """

    def expects(value: float) -> float:
        return 1 / (1 - (1 / value))

    values = range(5)
    mapped = list(result_mapper(values, invert, subtract_to_one_and_invert))
    assert mapped[2:5] == [Ok(expects(2)), Ok(expects(3)), Ok(expects(4))]
    assert not mapped[0]
    assert not mapped[1]
    assert type(mapped[0].error) is ZeroDivision
    assert type(mapped[1].error) is ZeroDivision

def test_result_reducer_ok_case() -> None:
    """
    Tests the result reducer
    """
    assert result_reducer(
        Ok(2), Ok(3), Ok(5),
        initializer=0,
        reducer=lambda a,b: a + b,
    ) ==  Ok(10)

    l = [1]
    assert result_reducer(
        Ok(2), Ok(3), Ok(5),
        initializer=l,
        reducer=lambda l, a: l + [a],
    ) ==  Ok([1, 2,3,5])

def test_result_reducer_err_case_case() -> None:
    """
    Tests the result reducer
    """
    assert result_reducer(
        Ok(2), Err(), Ok(5),
        initializer=0,
        reducer=lambda a,b: a + b,
    ) ==  Err()

def test_result_chain_ok_case() -> None:
    """
    Tests the result reducer
    """
    def get_number(number: int) -> Ok[int]:
        return Ok(number)

    assert chain_results(
        partial(get_number, 2),
        partial(get_number, 3), 
        partial(get_number, 5), 
        initializer=0,
        reducer=lambda a,b: a + b,
    ) ==  Ok(10)

    l = [1]

    assert chain_results(
        partial(get_number, 2),
        partial(get_number, 3), 
        partial(get_number, 5), 
        initializer=l,
        reducer=lambda l, a: l + [a],
    ) ==  Ok([1,2,3,5])

def test_result_chain_err_case() -> None:
    """
    Tests the result reducer
    """
    def get_number(number: int):
        return Ok(number)
    
    def do_err():
        return Err()
    
    # we want to make sure that panic is never run
    def panic():
        raise RuntimeError()

    assert chain_results(
        partial(get_number, 2),
        do_err,
        panic,
        initializer=0,
        reducer=lambda a,b: a + b,
    ) ==  Err()